from __future__ import print_function, division, absolute_import, unicode_literals

from time import sleep
import unittest
import datetime

from six.moves import range, zip

from examples import multistopwatch_examples
from bistiming import MultiStopwatch
from .utils import assert_timedelta_close_seconds_list, isclose


class TestMultiStopwatch(unittest.TestCase):
    def test_multistopwatch_examples(self):
        multistopwatch_examples.main()

    def test_multi_stopwatch(self):
        timers = MultiStopwatch(2, verbose=False)
        for _ in range(5):
            for _ in range(2):
                with timers[0]:
                    sleep(0.1)
            with timers[1]:
                sleep(0.1)

        # test get_cumulative_elapsed_time()
        assert_timedelta_close_seconds_list(
            timers.get_cumulative_elapsed_time(), [1, 0.5]
        )

        # test get_percentage()
        for p, exp in zip(timers.get_percentage(), [2 / 3, 1 / 3]):
            self.assertTrue(isclose(p, exp, rel_tol=0.05))

        # test get_n_splits()
        self.assertListEqual(timers.get_n_splits(), [10, 5])

        # test get_mean_per_split()
        assert_timedelta_close_seconds_list(timers.get_mean_per_split(), [0.1, 0.1])

        # test get_statistics()
        self.assertListEqual(
            sorted(timers.get_statistics().keys()),
            ["cumulative_elapsed_time", "mean_per_split", "n_splits", "percentage"],
        )

    def test_multi_stopwatch_without_elapsed_time(self):
        timers = MultiStopwatch(2, verbose=False)
        self.assertListEqual(
            timers.get_cumulative_elapsed_time(), [datetime.timedelta()] * 2
        )
        with self.assertRaises(ValueError):
            timers.get_percentage()
        self.assertListEqual(timers.get_n_splits(), [0, 0])
        self.assertListEqual(timers.get_mean_per_split(), [datetime.timedelta()] * 2)
        with self.assertRaises(ValueError):
            timers.get_statistics()

    def test_empty_multi_stopwatch(self):
        timers = MultiStopwatch()
        self.assertListEqual(timers.get_cumulative_elapsed_time(), [])
        with self.assertRaises(ValueError):
            timers.get_percentage()
        self.assertListEqual(timers.get_n_splits(), [])
        self.assertListEqual(timers.get_mean_per_split(), [])
        with self.assertRaises(ValueError):
            timers.get_statistics()



from __future__ import print_function, division, absolute_import, unicode_literals

from functools import partial
import logging
import datetime

import six


class Stopwatch(object):
    """A logging-friendly stopwatch with splitting function.

    Parameters
    ----------
    description : str
        The log to show at starting time (entering with-block or calling :meth:`start`)
        or ending time (exiting with-block or calling :meth:`split`).
    logger : :data:`~typing.Callable`
        A callable that accepts `logging_level` as its first argument and a :class:`str` to
        log as its first argument (basically, a :class:`logging.Logger` object). If `None`,
        use :func:`six.print_`, which is similar to the built-in :func:`print` in Python 3.
        When using with ``end_in_new_line=True``, it requires `end` and `flush` parameters.
    logging_level : int
        If `logger` is not `None`, this is the first argument to be passed to `logger`.
        Usually, this should be `logging.{DEBUG, INFO, WARNING, ERROR, CRITICAL}`.
        (default: `logging.INFO`)
    verbose_start : bool
        Wether to log at starting time (entering with-block or calling :meth:`start`).
    verbose_end : bool
        Wether to log at ending time (exiting with-block or calling :meth:`split`).
    end_in_new_line : bool
        Wether to log the ending log in a new line. If `False`, the starting log will
        not have a trailing new line, so the ending log can be logged in the same line.
        This requires `logger` to have `end` and `flush` parameters, or just
        ``logger=None``.
    prefix : str
        The prefix added to `description`.
    verbose : bool
        If `False`, turn off all the logs, that is, `verbose_start` and `verbose_end`
        will be set to `False`.

    Attributes
    ----------
    split_elapsed_time : List[datetime.timedelta]
        The elapsed time of each split (excluding the current split).
    """

    def __init__(
        self,
        description="",
        logger=None,
        logging_level=logging.INFO,
        verbose_start=True,
        verbose_end=True,
        end_in_new_line=True,
        prefix="...",
        verbose=True,
    ):
        if logger is not None:
            self.log = partial(logger.log, logging_level)
        else:
            self.log = six.print_
        self.description = prefix + description
        if verbose:
            self.verbose_start = verbose_start
            self.verbose_end = verbose_end
        else:
            self.verbose_start = False
            self.verbose_end = False
        self.end_in_new_line = end_in_new_line
        self.reset()

    def start(self, verbose=None, end_in_new_line=None):
        """Start the stopwatch if it is paused.

        If the stopwatch is already started, then nothing will happen.

        Parameters
        ----------
        verbose : Optional[bool]
            Wether to log. If `None`, use `verbose_start` set during initialization.
        end_in_new_line : Optional[bool]]
            If `False`, prevent logging the trailing new line. If `None`, use
            `end_in_new_line` set during initialization.
        """
        if self._start_time is not None and self._end_time is None:
            # the stopwatch is already running
            return self
        if verbose is None:
            verbose = self.verbose_start
        if verbose:
            if end_in_new_line is None:
                end_in_new_line = self.end_in_new_line
            if end_in_new_line:
                self.log(self.description)
            else:
                self.log(self.description, end="", flush=True)
        self._end_time = None
        self._start_time = datetime.datetime.now()
        return self

    def pause(self):
        """Pause the stopwatch.

        If the stopwatch is already paused, nothing will happen.
        """
        if self._end_time is not None:
            # the stopwatch is already paused
            return
        self._end_time = datetime.datetime.now()
        self._elapsed_time += self._end_time - self._start_time

    def get_elapsed_time(self):
        """Get the elapsed time of the current split."""
        if self._start_time is None or self._end_time is not None:
            # the stopwatch is paused
            return self._elapsed_time
        return self._elapsed_time + (datetime.datetime.now() - self._start_time)

    def get_cumulative_elapsed_time(self):
        """Get the cumulative elapsed time without considering splits."""
        return self._cumulative_elapsed_time + self.get_elapsed_time()

    def log_elapsed_time(self, prefix="Elapsed time: "):
        """Log the elapsed time of the current split.

        Parameters
        ----------
        prefix : str
            The prefix of the log.
        """
        self.log("{}{}".format(prefix, self.get_elapsed_time()))

    def split(
        self,
        verbose=None,
        end_in_new_line=None,
        message_format="done in {elapsed_time}",
    ):
        """Save the elapsed time of the current split and restart the stopwatch.

        The current elapsed time will be appended to :attr:`split_elapsed_time`.
        If the stopwatch is paused, then it will remain paused.
        Otherwise, it will continue running.

        Parameters
        ----------
        verbose : Optional[bool]
            Wether to log. If `None`, use `verbose_end` set during initialization.
        end_in_new_line : Optional[bool]]
            Wether to log the `description`. If `None`, use `end_in_new_line` set during
            initialization.
        message_format: str
            The string that will be formatted using ``message_format.format(...)``
            and be logged as the ending message.
            Available variables: `elapsed_time`.
        """
        elapsed_time = self.get_elapsed_time()
        self.split_elapsed_time.append(elapsed_time)
        self._cumulative_elapsed_time += elapsed_time
        self._elapsed_time = datetime.timedelta()
        if verbose is None:
            verbose = self.verbose_end
        if verbose:
            message = message_format.format(elapsed_time=elapsed_time)
            if end_in_new_line is None:
                end_in_new_line = self.end_in_new_line
            if end_in_new_line:
                self.log("{} {}".format(self.description, message))
            else:
                self.log(" {}".format(message))
        self._start_time = datetime.datetime.now()

    def reset(self):
        """Reset the stopwatch."""
        self._start_time = None
        self._end_time = None
        self._elapsed_time = datetime.timedelta()
        self._cumulative_elapsed_time = datetime.timedelta()
        self.split_elapsed_time = []

    def __enter__(self):
        """Call :meth:`start`."""
        return self.start()

    def __exit__(self, exc_type, exc, exc_tb):
        """Call :meth:`pause` and then :meth:`split`."""
        self.pause()
        if exc_type is None:
            self.split()
        else:
            self.split(message_format="got exception in {elapsed_time}")



from __future__ import print_function, division, absolute_import, unicode_literals

import datetime

from six.moves import UserList, range
from tabulate import tabulate

from . import Stopwatch
from .utils import div_timedelta


class MultiStopwatch(UserList):
    """Use multiple :class:`Stopwatch` to profile and compare multiple code segments.

    Because this class is inheritted from :class:`~collections.UserList`,
    it also supports all the methods in the built-in :class:`list`.

    Parameters
    ----------
    n : Union[int, Iterable, None]
        If `n` is an `int`, then initialize a `list` with `n` :class:`Stopwatch`.
        If `n` is `None`, then initialize an empty `list`.
        Otherwise, directly use `n` to initialize a `list`.
    *args
        Other arguments will be passed to initialize :class:`Stopwatch`.
    **kwargs
        Other keyword arguments will be passed to initialize :class:`Stopwatch`.
    """

    def __init__(self, n=None, *args, **kwargs):
        if isinstance(n, int):
            super(MultiStopwatch, self).__init__(
                Stopwatch(*args, **kwargs) for i in range(n)
            )
        else:
            super(MultiStopwatch, self).__init__(n)

    def get_cumulative_elapsed_time(self):
        """Get the cumulative elapsed time of each stopwatch (including the current split).

        Returns
        -------
        cumulative_elapsed_time : List[datetime.timedelta]
        """
        return [stopwatch.get_cumulative_elapsed_time() for stopwatch in self]

    def get_percentage(self):
        """Get the cumulative time percentage of each stopwatch (including the current split).

        Returns
        -------
        cumulative_elapsed_time_percentage : List[float]
        """
        cumulative_elapsed_time = self.get_cumulative_elapsed_time()
        sum_elapsed_time = sum(cumulative_elapsed_time, datetime.timedelta())
        if not sum_elapsed_time:
            raise ValueError("cannot get percentage if there is no any elapsed time")
        return [div_timedelta(t, sum_elapsed_time) for t in cumulative_elapsed_time]

    def get_n_splits(self):
        """Get number of splits of each stopwatch (excluding the current split).

        Returns
        -------
        n_splits : List[int]
        """
        return [len(stopwatch.split_elapsed_time) for stopwatch in self]

    def get_mean_per_split(self):
        """Get the mean elapsed time per split of each stopwatch (excluding the current split).

        Returns
        -------
        mean_elapsed_time_per_split : List[datetime.timedelta]
        """
        return [
            (
                div_timedelta(
                    sum(stopwatch.split_elapsed_time, datetime.timedelta()),
                    len(stopwatch.split_elapsed_time),
                )
                if stopwatch.split_elapsed_time
                else datetime.timedelta()
            )
            for stopwatch in self
        ]

    def get_statistics(self):
        """Get all statistics as a dictionary.

        Returns
        -------
        statistics : Dict[str, List]
        """
        return {
            "cumulative_elapsed_time": self.get_cumulative_elapsed_time(),
            "percentage": self.get_percentage(),
            "n_splits": self.get_n_splits(),
            "mean_per_split": self.get_mean_per_split(),
        }

    def format_statistics(self, tablefmt="fancy_grid"):
        """Format the statistics using tabulate.

        Parameters
        ----------
        tablefmt: str
            See the available options in
            `tabulate's documentation <https://github.com/astanin/python-tabulate#table-format>`_.
        """
        return tabulate(self.get_statistics(), headers="keys", tablefmt=tablefmt)



from __future__ import print_function, division, absolute_import, unicode_literals

from six.moves import zip


def assert_timedelta_close_seconds_list(d_list, s_list, epsilon=0.05):
    for d, s in zip(d_list, s_list):
        assert_timedelta_close_seconds(d, s, epsilon)


def assert_timedelta_close_seconds(d, s, epsilon=0.05):
    d_s = d.days * 86400 + d.seconds + d.microseconds * 0.000001
    percentage_error = abs(d_s - s) / s
    assert percentage_error < epsilon


def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)



from __future__ import print_function, division, absolute_import, unicode_literals

import datetime


def div_timedelta_int(d, i):
    d_us = d.microseconds + 1000000 * (d.seconds + 86400 * d.days)
    return datetime.timedelta(microseconds=d_us / i)


def div_timedelta(d1, d2):
    if isinstance(d2, int):
        return div_timedelta_int(d1, d2)
    d1_us = d1.microseconds + 1000000 * (d1.seconds + 86400 * d1.days)
    d2_us = d2.microseconds + 1000000 * (d2.seconds + 86400 * d2.days)
    return d1_us / d2_us



__all__ = ["Stopwatch", "MultiStopwatch"]

try:
    # for Python >= 3.8
    from importlib.metadata import version
except ImportError:
    # for Python < 3.8, the package importlib-metadata will be installed
    from importlib_metadata import version

__version__ = version("bistiming")

from .stopwatch import Stopwatch  # noqa: F401
from .multistopwatch import MultiStopwatch  # noqa: F401

SimpleTimer = Stopwatch  # backward-compatible to < 0.2



from __future__ import print_function, division, absolute_import, unicode_literals

import datetime
from time import sleep

from examples import stopwatch_examples
from bistiming import Stopwatch
from .utils import assert_timedelta_close_seconds


def test_stopwatch_examples():
    stopwatch_examples.main()


def test_low_level_api():
    timer = Stopwatch()
    sleep(0.1)
    assert timer.get_elapsed_time() == datetime.timedelta()
    timer.start()
    sleep(0.1)
    assert_timedelta_close_seconds(timer.get_elapsed_time(), 0.1)
    sleep(0.1)
    timer.pause()
    assert_timedelta_close_seconds(timer.get_elapsed_time(), 0.2)
    sleep(0.1)
    assert_timedelta_close_seconds(timer.get_elapsed_time(), 0.2)
    timer.split()  # 0:00:00.2
    assert timer.get_elapsed_time() == datetime.timedelta()
    assert_timedelta_close_seconds(timer.get_cumulative_elapsed_time(), 0.2)
    sleep(0.1)
    timer.start()
    sleep(0.1)
    assert_timedelta_close_seconds(timer.get_elapsed_time(), 0.1)
    assert_timedelta_close_seconds(timer.get_cumulative_elapsed_time(), 0.3)
    timer.split()  # 0:00:00.1
    sleep(0.1)
    timer.pause()
    timer.split()  # 0:00:00.1
    assert_timedelta_close_seconds(timer.get_cumulative_elapsed_time(), 0.4)
    assert_timedelta_close_seconds(timer.split_elapsed_time[0], 0.2)
    assert_timedelta_close_seconds(timer.split_elapsed_time[1], 0.1)
    assert_timedelta_close_seconds(timer.split_elapsed_time[2], 0.1)
    timer.reset()

    assert timer.get_elapsed_time() == datetime.timedelta()
    assert timer.get_cumulative_elapsed_time() == datetime.timedelta()
    assert timer.split_elapsed_time == []
    sleep(0.1)
    timer.start()
    sleep(0.1)
    assert_timedelta_close_seconds(timer.get_elapsed_time(), 0.1)
